import os
import xarray as xr
import arviz as az
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
import ipywidgets as widgets
from IPython.display import display
import glob


def read_log_file_as_dataframe(file_path,
                               cut_to_first=None):
    """Read a log file into a pandas DataFrame.

    Parameters
    ----------------
    file_path: str
        Path to the log file.
    cut_to_first : int, default None
        Remove Samples/links over this number.


    Returns
    -----------
    Pandas.DataFrame
    """
    trace_df = pd.read_table(file_path, sep='\t', comment='#')
    if cut_to_first is not None:
        trace_df = trace_df[trace_df['Sample'] <= cut_to_first]
    return trace_df

def read_log_files_as_xarraydata(log_files,
                                 start=0,
                                 cut_to_first=None):
    """
    Reads log files from a list and returns a xarray DataArray.

    Parameters
    ----------
    log_files: list, tuple or dict  of str
        Contains paths of log files. If list or tuple chains are indexed from
         incrementally start argument. If dictionary keys are used as the chain index.
    start : int
        Starting index of the first log file to read.
        Remove Samples/links over this number in log files.
    cut_to_first : int, default None
        Remove Samples/links over this number.

    Returns
    -------
    xarray.DataArray
    """
    dfs = []
    if isinstance(log_files, (list,tuple)):
        for i, log_file in enumerate(log_files, start=start):
            df = read_log_file_as_dataframe(log_file, cut_to_first=cut_to_first)
            df['chain'] = i
            dfs.append(df)
    elif isinstance(log_files, dict):
        for i, log_file in log_files.items():
            df = read_log_file_as_dataframe(log_file, cut_to_first=cut_to_first)
            df['chain'] = i
            dfs.append(df)
    else:
        raise TypeError('log_files must be a list, tuple or a dictionary.')

    df = pd.concat(dfs)
    df.rename(columns={'Sample': 'draw'}, inplace=True)
    df.set_index(["chain", "draw"], inplace=True)
    return xr.Dataset.from_dataframe(df)


def read_log_files_as_posterior(log_files,
                                start=0,
                                cut_to_first=None):
    """
    Reads log files from a list and returns an arviz BEASTDiag Inference  DataArray.
    Parameters
    ----------
    log_files: list of str
        List of log files.
    start : int
        Starting index of the first log file to read.
            sample_name: str, default='Sample'
    cut_to_first : int, default None
        Remove Samples/links over this number in log files.

    Returns
    -------
    arviz.InferenceData
        An arviz BEASTDiag Inference  DataArray
    """
    xdata = read_log_files_as_xarraydata(log_files,
                                         start=start,
                                         cut_to_first=cut_to_first)
    return az.InferenceData(posterior=xdata)


def burnin_posterior(posterior,
                     proportion=None,
                     percentage=None,
                     number=None,
                     sample_name='draw'):
    """
    Perform a burn-in on a posterior.

    Parameters
    ----------
    posterior: arviz.data.inference_data.InferenceData
        DataArray with posterior. Must have 'chain' and 'draw' dimension names for in
         posterior.
    proportion: float, default=None
        Proportion of posterior to burn-in.
    percentage: float, default=None
         Percentage of posterior to burn-in.
    number: int
        Number of burn-in points.
    sample_name: str, default='draw'
        Name of dimension of xml_set.

    Returns
    -------
    arviz.data.inference_data.InferenceData
    """
    if proportion is None and number is None and percentage is None:
        raise ValueError("Either proportion, percentage or number must be provided")
    if proportion is not None:
        if not isinstance(proportion, float):
            raise TypeError("Proportion must be a float")
        if number is not None or percentage is not None:
            raise ValueError("One of proportion, percentage or number must be provided")
        number = round(proportion * len(posterior.posterior[sample_name]))
    elif percentage is not None:
        if not isinstance(percentage, (int,float)):
            raise TypeError("Percentage must be a float or an integer.")
        if number is not None or proportion is not None:
            raise ValueError("One of proportion, percentage or number must be provided")
        number = round(percentage/100 * len(posterior.posterior[sample_name]))
    elif isinstance(number, int):
        raise TypeError("Number must be an integer")
    selection = {sample_name: slice(number, None)}
    return posterior.isel(**selection, groups="posterior")


def select_chains_from_posterior(posterior, selection):
    """
    Select chains from a posterior.
    Parameters
    ----------
    posterior: arviz.data.inference_data.InferenceData
        DataArray with posterior. Must have 'chain' and 'draw' dimension names for in
         posterior.
    selection: list of strings or ints
        List of chains to select from posterior.

    Returns
    -------
    arviz.data.inference_data.InferenceData
    """
    selection = {'chain': selection}
    return posterior.sel(**selection, groups="posterior")


def plot_traces(posterior, parameters, labels=None, legend=True):
    """
    Plot traces (a wrapper for arviz.plot_trace with a better positioned legend).
    
    Parameters
    ----------
    posterior: arviz.data.inference_data.InferenceData
        DataArray with posterior. Must have 'chain' and 'draw' dimension names for in
         posterior.
    parameters: list of str
        List of parameters names.
    labels: list, default=None
        List of labels to use for legend.
    legend: bool, default=True
        Add legend.

    Returns
    -------
    Numpy array of matplotlib axes.
    """
    num_params = len(parameters)
    fig, axs = plt.subplots(nrows=num_params, ncols=2, figsize=(13, 2*num_params))
    if num_params == 1:
        axs = axs.reshape((1, 2))
    plt.subplots_adjust(hspace=0.4)
    traces = az.plot_trace(posterior,
                           axes=axs,
                           var_names=parameters,
                           chain_prop="color",
                           compact=True,
                           legend=legend)
    if legend:
        if labels is not None:
             axs[0][1].legend(labels=labels)
        sns.move_legend(axs[0][1], loc="upper left", bbox_to_anchor=(1, 1), ncol=1)
    for i, parameter in enumerate(parameters):
        axs[i][0].set_title(parameter, x=1.1)
        axs[i][1].set_title('')
        axs[i][1].ticklabel_format(style='sci', axis='x', scilimits=(0, 0))
    return fig, axs


class BEASTDiag:
    """
    Class to perform a burn-in and chain selection on a sampled posterior from BEAST.

    Attributes
    ---------------
    directory:  str
        Directory where log and trees files are located.
    original_posterior: arviz.data.inference_data.InferenceData
        Posterior with no chains removed and no burn-in.
    parameters: list of str
        List of parameters names.
    parameters_dict: dict {str: [strs]}
        Dictionary of parameters categories.
    original_chains: list of strs
         Chains in original posterior.
    selected_posterior: arviz.data.inference_data.InferenceData
        Posterior with chains removed and burn-in applied.
    burinin_percentage: int
        Percentage burn-in applied to self.selected_posterior.
    selected_chains: list of strs
        Chains in selected posterior.
    diagnostics_of_selection: pandas.DataFrame
        Summary of diagnostics of self.selected_posterior. Product of
        arviz.summary(self.selected_posterior, kind='diagnostics')

    """
    def __init__(self,
                 directory,
                 start=0,
                 cut_to_first=None,
                 ignore_iqtree_logfiles=True):
        """

        Parameters
        ----------
        directory:  str
            Directory where log and trees files are located.
        start : int
            Starting index of the first log file to read.
                sample_name: str, default='Sample'
        cut_to_first : int, default None
            Remove Samples/links over this number in log files.
        """
        self.directory = directory
        if not os.path.exists(directory):
            raise FileNotFoundError('directory does not exist')
            
        if ignore_iqtree_logfiles:
            log_paths = {
                path.removeprefix(directory + '/').removesuffix('.log'): path
                for path in glob.glob(directory + '/*.log')
                if not path.endswith('iqtree.log')
            }
        else:
            log_paths = {
                path.removeprefix(directory + '/').removesuffix('.log'): path
                for path in glob.glob(directory + '/*.log')
            }
        chains = list(log_paths.keys())
        self.original_posterior = read_log_files_as_posterior(log_paths,
                                                              start=start,
                                                              cut_to_first=cut_to_first)

        self.parameters = list(self.original_posterior.posterior.data_vars)
        self.parameters_dict = None
        self.original_chains = sorted(chains)
        self.selected_posterior = None
        self.burinin_percentage = 0
        self.selected_chains = chains
        self.diagnostics_of_selection = None
        self.cut_to_first = cut_to_first

    def set_burnin(self, percentage=10, sample_name='draw'):
        """
        Perform a burn-in on a posterior.

        Parameters
        ----------------
        percentage: int, default=10
             Percentage of posterior to burn-in.
        sample_name: str, default='draw'
            Name of dimension of xml_set.

        """
        if not isinstance(percentage, int):
            raise TypeError('percentage must be an integer')
        self.burinin_percentage = percentage
        selected_posterior = burnin_posterior(posterior=self.original_posterior,
                                              percentage=self.burinin_percentage,
                                              sample_name=sample_name)
        self.selected_posterior = select_chains_from_posterior(selected_posterior,
                                                          self.selected_chains)




    def select_chains(self, chains=None, **kwargs):
        """
        Select chains to use in posterior.

        Parameters
        ----------------
        chains: list of strings or ints, default=None
            List of chains to select from posterior.
        kwargs: dict, default=None
            Chains labels with bool value. If True chain is included in posterior.
            If False chain is excluded from posterior.

        """
        if chains is None and not kwargs:
            raise ValueError("Either chains or kwargs must be provided")
        if chains is not None and kwargs:
            raise ValueError("Either chains or kwargs must be provided")

        if kwargs:
            for chain, value in kwargs.items():
                if chain not in self.original_chains:
                    raise ValueError("Chain {0} not found in original posterior, ".format(chain) +
                                     'see self.original_chains.')
                if not isinstance(value, bool):
                    raise TypeError("Chain {0} was not given a boolean value.".format(chain))
            for chain in self.original_chains:
                if chain not in kwargs:
                    raise ValueError("Chain {0} not found in kwargs, ".format(chain))
            self.selected_chains = [chain for chain in self.original_chains if kwargs[chain]]

        if chains is not None:
            for chain in chains:
                if chain not in self.original_chains:
                    raise ValueError("Chain {0} not found in original posterior, ".format(chain) +
                                     'see self.original_chains.')
            self.selected_chains = chains

        selected_posterior = burnin_posterior(posterior=self.original_posterior,
                                              percentage=self.burinin_percentage)
        self.selected_posterior = select_chains_from_posterior(selected_posterior, self.selected_chains)


    def diagnose_selection(self):
        """
        Diagnose selected_posterior.

        Updates self.diagnosis_of_selection via
        arviz.summary(self.selected_posterior, kind='diagnostics')

        """
        self.diagnosis_of_selection = az.summary(self.selected_posterior,
                                                                  kind='diagnostics')

    def merge_logs_to_csv(self, output_file='merged_log.csv', like_logcombiner=True):
        """
        Merge selected log files into one csv file.

        Parameters
        ----------
        output_file : str, default='merged.log'
            Name of output file. Saved in self.directory.
        like_logcombiner : bool, default=True
            Merged output looks similar to merged output from logcombiner.

        """
        df = self.selected_posterior.to_dataframe()
        if like_logcombiner:
            step = df.iloc[1, 1] - df.iloc[0, 1]
            df[df.columns[1]] = range(0, len(df) * step, step)
            if df.columns[1] != 'Sample':
                relabel_dict = {df.columns[1]: 'Sample'}
                df.rename(columns=relabel_dict, inplace=True)
            df = df.drop(columns=['chain'])
        df.to_csv(f'{self.directory}/{output_file}', index=False)

    def logcombiner_args(self, output_pretfix, output_file='merged', suffix='.log'):
        """
        Create logcombiner args for merging selected chain files (.log or .trees) into one.

        Parameters
        ----------
        output_pretfix: str
            Prefix for output file. Saved in self.directory.
        output_file : str, default='merged'
            Name of output file. Saved in self.directory.
        suffix: str, default '.log'
            Suffix of files to merge.

        Returns
        -------
        lc_args: dict {str: str}
            Logcombiner args.
        """
        if self.cut_to_first is not None:
            raise AttributeError(
                "Use of cut_to_first and BEASTs logcombiner is not yet implemented." +
                'However, merge_logs_to_csv method of this class can perform your' +
                ' selection and produce a merged csv from the log files.')
        if suffix not in ['.log', '.trees']:
            raise ValueError("suffix must be either '.log', '.trees'.")
        output_file = f'{output_pretfix}{output_file}{suffix}'
        selected_files = [f"{self.directory}/{chain}{suffix}"
                              for chain in self.selected_chains]
        name_pre_fix = suffix.replace('.', '')
        lc_args = {
            f'{name_pre_fix}_file_burnin':  self.burinin_percentage,
            f'{name_pre_fix}_file_output': output_file,
            f'{name_pre_fix}_files_to_combine': ' '.join(selected_files)
        }
        return lc_args

    def merging_outputs_params(self, output_path, kernel_name = 'beast_pype', xml_set=None):
        if xml_set is None:
            output_prefix = f'{output_path}/'
        else:
            output_prefix = f'{output_path}/{xml_set}_'
        params = {
            **self.logcombiner_args(output_prefix, suffix='.log'),
            **self.logcombiner_args(output_prefix, suffix='.trees')
        }
        params['kernel_name'] = kernel_name
        return params


    def _widget_interaction(self, percentage, parameters, **kwargs):
        for chain, value in kwargs.items():
            if chain not in self.original_chains:
                raise ValueError("Chain {0} not found in original posterior, ".format(chain) +
                                 'see self.original_chains.')
            if not isinstance(value, bool):
                raise TypeError("Chain {0} was not given a boolean value.".format(chain))
        for chain in self.original_chains:
            if chain not in kwargs:
                raise ValueError("Chain {0} not found in kwargs, ".format(chain))

        selected_chains = [chain for chain in self.original_chains if kwargs[chain]]
        posterior_modified = False
        if selected_chains != self.selected_chains:
            posterior_modified = True
            self.select_chains(selected_chains)
        if percentage != self.burinin_percentage:
            posterior_modified = True
            self.set_burnin(percentage=percentage)
        if posterior_modified:
            self.diagnose_selection()
        if len(selected_chains) > 0:
            stats, trace = self._display_diagnosis(parameters)
            plt.show()
            display(stats)

    @property
    def parameters_types(self):
        return self.parameters_dict.keys()

    def _display_diagnosis(self, parameters):
        if self.selected_posterior is None:
            raise AssertionError('BEASTDiag must be modified first. ' +
                                 'Use method set_burnin_and_chains().')
        if isinstance(parameters, str) and parameters in self.parameters_dict:
            parameters = self.parameters_dict[parameters]
        else:
            parameters = [parameters]
        for parameter in parameters:
            if parameter not in self.parameters:
                raise ValueError("Parameter {0} not found in posteriors list of parameters.".format(parameter))

        traces_fig, traces_ax = plot_traces(posterior=self.selected_posterior,
                                            parameters=parameters,
                                            labels=self.selected_chains)
        return self.diagnosis_of_selection.loc[parameters], traces_fig

    def clear_all_chains(self):
        """Clear all chains."""
        for chain in self.chain_checks:
            chain.value = False

    def select_all_chains(self):
        """Select all chains."""
        for chain in self.chain_checks:
            chain.value = True

    def generate_widget(self,
                 parameters_displayed=4):
        """
        Generates widget for selecting burn-in and chains, for use in Jupyter notebook.

        Parameters
        -------------
        parameters_displayed: int
            Number of parameters to display at a time.

        Returns
        -----------
        ipywidgets.widgets.VBox

        """
        num_params = len(self.parameters)
        parameters_dict = {}
        from_index = 0
        to_index = parameters_displayed
        parameter_set = 1
        while to_index <= num_params-1:
            parameters_dict[f'parameters set {str(parameter_set)}'] = self.parameters[from_index:to_index]
            parameter_set +=1
            from_index += parameters_displayed
            to_index += parameters_displayed
        parameters_dict[f'parameters set {str(parameter_set)}'] = self.parameters[from_index:]
        self.parameters_dict = parameters_dict
        burnin_title = widgets.HTML('Burn-in')
        burnin_selector = widgets.IntSlider(
            value=10,
            min=0,
            max=100,
            step=1,
            description='%:',
            disabled=False,
            continuous_update=False,
            orientation='horizontal',
            readout=True
        )
        chain_title = widgets.HTML('Chains')
        self.chain_checks = []
        for chain in self.original_chains:
            self.chain_checks.append(widgets.Checkbox(value=True, description=chain, disabled=False))
        chain_selector = widgets.GridBox(self.chain_checks,
                                         layout=widgets.Layout(grid_template_columns="repeat(4, 250px)"))
        # select_all_buttion = widgets.Button(description='Select all Chains',
        #                                  tooltip='Select all Chains',
        #                                  button_style='')
        # select_all_buttion.on_click(self.select_all_chains)
        # clear_all_buttion = widgets.Button(description='Clear all chains',
        #                                  tooltip='Clear all Chains',
        #                                  button_style='')
        # clear_all_buttion.on_click(self.clear_all_chains)
        # select_clear_all = widgets.HBox(children=[select_all_buttion, clear_all_buttion])
        burnin_and_chain_selector = widgets.VBox(children=[
            burnin_title,
            burnin_selector,
            chain_title,
            chain_selector,
            # select_clear_all
            ],
            titles=('Burn-in', 'Chains'))
        parameter_selector = widgets.Dropdown(options=self.parameters_types,
                                              description='Parameters:')
        output_widget = widgets.interactive_output(self._widget_interaction,
                                                   controls={
                                                       'percentage': burnin_selector,
                                                       'parameters': parameter_selector,
                                                       **{chain.description: chain for chain in self.chain_checks}
                                                   })
        beast_diag_widget = widgets.VBox([
            burnin_and_chain_selector,
            parameter_selector,
            output_widget])

        return beast_diag_widget


import pandas as pd
from beast_pype.outputs import read_log_file
import nbformat as nbf
import re
import os
from beast_pype.path_utils import path_to_workflow_modules


def add_unreported_outputs(notebook_template_path,
                           merged_logs_path,
                           output_path,
                           xml_set_comparisons=False,
                           as_version=4):
    """
    Add sections for unreported outputs to a notebook.

    Parameters
    ----------
    notebook_template_path: str
        Path to the notebook template. Metadata of the notebook will be searched for
        the entry "BEAST outputs reported". Any parameter not listed but occurring
        in the file at merged_log_path will then be added to the notebook.
    merged_logs_path: str
        Path to directory containing merged log files you wish to report on.
    output_path: str
        Path to save new report notebook file
    xml_set_comparisons: bool, default False
        Are the sections to be added for xml_set comparisons.
    as_version: int
        Ipnyb notebook version.

    Returns
    -------
    None
    """
    report = nbf.read(notebook_template_path, as_version=as_version)
    columns_already_reported = ['xml set', 'Sample']
    columns_already_reported += report.metadata["BEAST outputs reported"]['parameters']
    merged_log_paths = [file for file in os.listdir(merged_logs_path) if file.endswith('merged.log')]
    merged_log_columns = []
    for merged_log_path in merged_log_paths:
        if merged_log_path.endswith('.log'):
            merged_log_df = read_log_file(f'{merged_logs_path}/{merged_log_path}')
        elif merged_log_path.endswith('.csv'):
            merged_log_df = pd.read_csv(f'{merged_logs_path}/{merged_log_path}')
        elif merged_log_path.endswith('.tsv'):
            merged_log_df = pd.read_csv(f'{merged_logs_path}/{merged_log_path}', sep='\t')
        else:
            raise ValueError(f'Only csv, tsv or log files are supported for merged_log_path ({merged_logs_path}/{merged_log_path}).')
        columns_already_reported += [
            col for col in merged_log_df.columns
            if col.startswith(('Unnamed',
                *report.metadata["BEAST outputs reported"]["parameters starting with"]))
        ]
        merged_log_columns += merged_log_df.columns.to_list()
    columns_to_report = set(merged_log_columns) - set(columns_already_reported)
    columns_to_report = sorted(columns_to_report)
    for column in columns_to_report:
        report['cells'].append(nbf.v4.new_markdown_cell(f"## {column}"))
        variable_pre_fix = re.sub('[^a-zA-Z0-9]', '_', column)
        if xml_set_comparisons:
            report['cells'].append(
                nbf.v4.new_code_cell(
                    f"{variable_pre_fix}_ax = plot_comparative_box_violin(df_melted_for_seaborn, '{column}', xml_set_label=xml_set_label)\n" +
                    f"{variable_pre_fix}_hdi_df = hdi_pivot(df, '{column}', xml_set_label=xml_set_label)\n" +
                    f"display({variable_pre_fix}_hdi_df)")
            )
        else:
            report['cells'].append(
                nbf.v4.new_code_cell(
                    f"{variable_pre_fix}_fig, ax, {variable_pre_fix}_hdi_est ="+
                    f"plot_hist_kde(trace_df=trace_df, parameter='{column}', hdi_prob=0.95)\n"+
                    f"display({variable_pre_fix}_hdi_est)")
            )
    with open(output_path, 'w') as f:
        nbf.write(report, f)


def _update_notebook_metadata(notebook_path, metadata_update, output_path = None, as_version=4):
    notebook = nbf.read(notebook_path, as_version=as_version)
    notebook.metadata.update(metadata_update)
    if output_path is None:
        output_path = notebook_path
    with open(output_path , 'w') as f:
        nbf.write(notebook, f)

def gen_mcc_notebook(merged_trees_path, output_path, kernel_name='beast_pype', as_version=4):
    workflow_modules_path = path_to_workflow_modules()
    notebook = nbf.read(f"{workflow_modules_path}/Phase-5ii-Gen-MCC-Trees.ipynb", as_version=as_version)
    notebook['cells'].append(nbf.v4.new_code_cell(
        f'source activate {kernel_name}'
    ))
    merged_tree_paths = [file for file in os.listdir(merged_trees_path) if file.endswith('merged.trees')]
    for merged_tree_path in merged_tree_paths:
        prefix = merged_tree_path.replace('merged.trees', '')
        if prefix != '':
            notebook['cells'].append(nbf.v4.new_markdown_cell(
                f"## Producing {prefix.replace('_','')} MCC tree."
            ))
        notebook['cells'].append(nbf.v4.new_code_cell(
            f"treeannotator -burnin 0 {merged_trees_path}/{prefix}merged.trees {merged_trees_path}/{prefix}mcc_tree.nexus"
        ))
    with open(output_path , 'w') as f:
        nbf.write(notebook, f)


from base_test_classes import SimpleWorkflowVariationTest

xml_generation_notebook = 'Phase-3-Gen-BDSKY-Serial-xml.ipynb'
workflow = 'BDSKY-Serial'

class TestBDSKYSerialFull(SimpleWorkflowVariationTest):
    name = 'Test_BDSKY-Serial_Full'
    parameters_path = 'tests/parameters/BDSKY-Serial_full.yml'
    workflow = workflow
    initial_tree_building = True
    downsampling = True
    xml_generation = True
    xml_generation_notebook = xml_generation_notebook

class TestBDSKYSerialNoDownsampling(SimpleWorkflowVariationTest):
    name = 'Test_BDSKY-Serial_No_Downsampling'
    parameters_path = 'tests/parameters/BDSKY-Serial_no_downsampling.yml'
    workflow = workflow
    initial_tree_building = True
    xml_generation = True
    xml_generation_notebook = xml_generation_notebook


class TestBDSKYSerialPartitions(SimpleWorkflowVariationTest):
    name = 'Test_BDSKY-Serial_Partitions'
    parameters_path = 'tests/parameters/BDSKY-Serial_partitions.yml'
    workflow = workflow
    initial_tree_building = True
    xml_generation = True
    xml_generation_notebook = xml_generation_notebook

class TestBDSKYSerialNoInitialTree(SimpleWorkflowVariationTest):
    name = 'Test_BDSKY-Serial_No_Initial_Tree'
    parameters_path = 'tests/parameters/BDSKY-Serial_no-initial-tree.yml'
    workflow = workflow
    xml_generation = True
    xml_generation_notebook = xml_generation_notebook

class TestBDSKYSerialXmlReadyToGo(SimpleWorkflowVariationTest):
    name = 'Test_BDSKY-Serial_XML_Ready_To_Go'
    parameters_path = 'tests/parameters/BDSKY-Serial_xml-ready-to-go.yml'
    workflow = workflow
    xml_generation_notebook = xml_generation_notebook


"""Functions for creating dataframes of gamma distribution and making gama distribution related calculations."""

from scipy.stats import gamma
import seaborn as sns
import numpy as np
import pandas as pd


## Gamma

def gamma_from_mean(mu, sd=None, variance=None):
    """Convert mean and either sd or variance to shape (a) and scale of gama distribution.

    See:
        Bolker, Benjamin M. 2008. “Gamma.” In Ecological Models and Data in R, 131–133. Princeton University Press.
        https://en.wikipedia.org/wiki/Gamma_distribution#Maximum_likelihood_estimation
        https://stats.stackexchange.com/questions/342639/how-to-find-alpha-and-beta-from-a-gamma-distribution


    Parameters
    ----------
    mu : int or float
        Mean
    sd : int or float, optional
        Standard Deviation, by default None. If variance is None value is assumed to be 1.
    variance : int or float, optional
        Variance, by default None. If sd is None value is assumed to be 1.

    Returns
    -------
    dict {'a' : float, 'scale': float}
       Diction containing shape (a) and scale of gama distribution.
    """
    if variance is None:
        if sd is None:
            variance = 1
        else:
            variance = sd ** 2
    return {'a': (mu ** 2) / variance, 'scale': variance / mu}


def mean_rate_years_gamma_df(mu, sd=None, variance=None, loc=0, min_range=0.1, max_range=500, step=0.1,  return_gamma_dict=False):
    """Generate dataframe outlining pdf of gamma distribution for mean yearly rate.

    Parameters
    ----------
    mu : float or int
        Mean in years
    sd : float or int, optional
        Standard deviation in years., by default None
    variance : float or int, optional
        Variance in years, by default None
    loc : float or int, optional
        Location, by default 0.
    min_range : int, optional
        Min range value in years, by default 0.1
    max_range : int, optional
        Max range value in years, by default 100
    step : int, optional
        Step in years, by default 0.1
    return_gamma_dict: bool, default False
        Whether to return gamma distribution as a dictionary.


    Returns
    -------
    gamma_dict: dict {'a' : float, 'scale': float}
        Diction containing shape (a) and scale of gama distribution.
    traces_df: pandas.DataFrame
        Dataframe outlining pdf of gamma distribution for mean yearly rate.
    """

    gamma_dict = gamma_from_mean(mu=mu, sd=sd, variance=variance)
    gamma_dict['loc'] = loc
    df = gamma_df(**gamma_dict, min_range=min_range, max_range=max_range, step=step)
    if return_gamma_dict:
        return gamma_dict, df
    else:
        return df

def gamma_df(a, loc=0, scale=1, min_range=0.1, max_range=500, step=0.1):
    """Generate dataframe outlining pdf of gamma distribution showing yearly rate converted to daily rate and day period.


    Parameters
    ----------
    a : float or int
        Shape
    loc : float or int, optional
        Location, by default 0.
    scale : float
        Scale, by default 1.
    min_range : int, optional
        Min range value in years, by default 0.1
    max_range : int, optional
        Max range value in years, by default 100
    step : int, optional
        Step in years, by default 0.1

    Returns
    -------
    pandas.DataFrame


    """
    rate_years = np.arange(min_range, max_range + step, step=step)
    rate_days = rate_years / 365
    period_days = 1 / rate_days
    probabilities = gamma.pdf(x=rate_years, a=a, loc=loc, scale=scale)
    return pd.DataFrame({'Rate per Year': rate_years, 'Rate per Day':  rate_days,
                         'Period in Days': period_days,
                         'Probability': probabilities})

def plot_gamma_df(gamma_df):
    """Plot dataframe resulting from gamma_df function.

    Parameters
    ----------
    gamma_df : pandas.DataFrame

    Returns
    -------
    seaborn.lineplot

    """
    df_melt = pd.melt(gamma_df, id_vars=['Probability'])
    df_melt_mod = df_melt[df_melt['Probability'] > 0.0001]
    fig = sns.relplot(data=df_melt_mod, x="value", y="Probability",
                      col="variable", col_order=['Period in Days', 'Rate per Day', 'Rate per Year'],
                      kind='line',
                      facet_kws={'sharex': False})
    fig.set_titles(col_template="")
    fig.axes[0, 0].set_xlabel('Infectious Period in Days')
    fig.axes[0, 1].set_xlabel('Rate of Becoming Uninfectious per Day')
    fig.axes[0, 2].set_xlabel('Rate of Becoming Uninfectious per Year')
    return fig

## LogNormal


## Beta





from subprocess import run
import io
import os
import pandas as pd


def _extracting_value_from_string(string, val_name):
    string = string.replace(val_name, '').strip()

    def is_part_of_number(character):
        if character.isdigit() or character == '.':
            return True
        else:
            return False

    return float(''.join(filter(is_part_of_number, string)))


def _extracting_values_from_txt_file(file_path, val_names):
    if isinstance(val_names, str):
        val_names = [val_names]
    file = open(file_path, "r")
    data = {}

    for line in file:
        for val_name in val_names:
            if line.startswith(val_name):
                if val_name in data:
                    raise AssertionError(val_name + ' is listed more than once in file.')
                data[val_name] = _extracting_value_from_string(line, val_name)

    return data


def get_beast_runtimes(directory,
                       outfile_startswith='run-with-seed-',
                       outfile_endswith='.out'):
    """
    Extract beast runtimes from out files.

    Parameters
    ----------
    directory: str
        Path to the directory containing the out files.
    outfile_startswith: str
        Out file ending suffix.
    outfile_endswith
        Out file prefix.

    Returns
    ---------
    pandas.DataFrame

    """
    time_records = []
    for file in os.listdir(directory):
        if file.startswith(outfile_startswith) and file.endswith(outfile_endswith):
            entry = {
                'BEAST_run': file.replace('.out', ''),
                **_extracting_values_from_txt_file(os.path.join(directory, file), 'Total calculation time: ')
            }
            time_records.append(entry)

    runtimes = pd.DataFrame(time_records)
    runtimes['run_time_D_H_M_S'] = pd.to_timedelta(runtimes['Total calculation time: '], unit='s')
    runtimes.rename(columns={'Total calculation time: ': 'run_time_seconds'}, inplace=True)
    return runtimes


def get_slurm_job_stats(job_ids):
    """
    Get statistics on Slurm jobs.

    Parameters
    ----------
    job_ids: list of strs
        List of job ids.

    Returns
    -------
    job_stats: pd.DataFrame
        DataFrame of slurm job statistics.

    """
    job_ids_request = ','.join([f"{entry}.batch" for entry in job_ids])
    request = f"sacct --jobs={job_ids_request} --format=JobID,AllocTres,Elapsed,CPUTime,TotalCPU,MaxRSS -p --delimiter='/t'"
    results = run(request, shell=True, capture_output=True, text=True)
    run_info_batch = pd.read_csv(io.StringIO(results.stdout), sep='/t')
    run_info_batch = run_info_batch.loc[:, ~run_info_batch.columns.str.startswith('Unnamed')]
    run_info_batch['JobID'] = run_info_batch['JobID'].str.replace('.batch', '')
    suffix_to_scientific = {'K': 'e3', 'M': 'e6', 'G': 'e9', 'T': 'e12'}
    maxrss = run_info_batch['MaxRSS']
    for suffix, scientific in suffix_to_scientific.items():
        maxrss = maxrss.str.replace(suffix, scientific)
    run_info_batch['Max RAM Used (GB)'] = maxrss.astype(float) / 1e9
    run_info_batch[['Allocated CPUs', 'Allocated RAM (GB)', 'Allocated Nodes']] = run_info_batch['AllocTRES'].str.split(
        ',', expand=True)
    run_info_batch['Allocated CPUs'] = run_info_batch['Allocated CPUs'].str.replace('cpu=', '').astype(int)
    run_info_batch['Allocated Nodes'] = run_info_batch['Allocated Nodes'].str.replace('node=', '').astype(int)
    run_info_batch['Allocated RAM (GB)'] = run_info_batch['Allocated RAM (GB)'].str.replace('mem=', '')
    run_info_batch['Allocated RAM (GB)'] = run_info_batch['Allocated RAM (GB)'].str.replace('G', '').astype(float)
    run_info_batch['Elapsed'] = pd.to_timedelta(run_info_batch['Elapsed'].str.replace('-', ' days '))
    run_info_batch['CPUTime'] = pd.to_timedelta(run_info_batch['CPUTime'].str.replace('-', ' days '))
    for i, val in enumerate(run_info_batch['TotalCPU'].str.count(':') == 1):
        if val:
            run_info_batch.loc[i, 'TotalCPU'] = '00:' + run_info_batch.loc[i, 'TotalCPU']
    if any(run_info_batch['TotalCPU'].str.contains('-', regex=False)):
        run_info_batch['TotalCPU'].str.replace('-', ' days ')
    run_info_batch['TotalCPU'] = pd.to_timedelta(run_info_batch['TotalCPU'])
    run_info_batch['CPU Efficiency (%)'] = 100 * run_info_batch['TotalCPU'] / run_info_batch['CPUTime']
    run_info_batch['RAM Efficiency (%)'] = 100 * run_info_batch['Max RAM Used (GB)'] / run_info_batch[
        'Allocated RAM (GB)']

    job_ids_request = ','.join(job_ids)
    request = f"sacct --jobs={job_ids_request} --format=JobID,JobName,Timelimit -p --delimiter='/t'"
    results = run(request, shell=True, capture_output=True, text=True)
    run_info = pd.read_csv(io.StringIO(results.stdout), sep='/t')
    run_info = run_info[run_info['JobID'].isin(job_ids)]
    run_info = run_info.loc[:, ~run_info.columns.str.startswith('Unnamed')]
    run_info['Timelimit'] = pd.to_timedelta(run_info['Timelimit'].str.replace('-', ' days '))

    job_stats = run_info.merge(run_info_batch, on='JobID')
    job_stats['Timelimit Used %'] = 100 * job_stats['Elapsed'] / job_stats['Timelimit']
    job_stats = job_stats[['JobID', 'JobName', 'Elapsed', 'Timelimit', 'Timelimit Used %', 'Allocated Nodes',
                           'Allocated CPUs', 'TotalCPU', 'CPUTime', 'CPU Efficiency (%)', 'Max RAM Used (GB)',
                           'Allocated RAM (GB)', 'RAM Efficiency (%)']]
    return job_stats


from base_test_classes import ComparativeWorkflowVariationTest

xml_generation_notebook = 'Phase-3-Gen-Generic-xml.ipynb'
workflow = 'Generic-Comparative'

class TestGenericComparativeFull(ComparativeWorkflowVariationTest):
    name = 'Test_Generic_Comparative_Full'
    parameters_path = 'tests/parameters/Generic-Comparative_full.yml'
    workflow = workflow
    initial_tree_building = True
    downsampling = True
    xml_generation = True
    xml_generation_notebook = xml_generation_notebook

class TestGenericComparativeNoDownsampling(ComparativeWorkflowVariationTest):
    name = 'Test_Generic_Comparative_No_Downsampling'
    parameters_path = 'tests/parameters/Generic-Comparative_no_downsampling.yml'
    workflow = workflow
    initial_tree_building = True
    xml_generation = True
    xml_generation_notebook = xml_generation_notebook

class TestGenericComparativeNoInitialTree(ComparativeWorkflowVariationTest):
    name = 'Test_Generic_Comparative_No_Initial_Tree'
    parameters_path = 'tests/parameters/Generic-Comparative_no-initial-tree.yml'
    workflow = workflow
    xml_generation = True
    xml_generation_notebook = xml_generation_notebook


import click
import os
from beast_pype.nb_utils import execute_notebook
from beast_pype.path_utils import path_to_workflows, path_to_report_templates
from beast_pype.diagnostics import gen_beast_diagnostic_nb
from datetime import datetime
from papermill.iorw import read_yaml_file


workflows_path = path_to_workflows()
available_workflows = [file.name for file in workflows_path.glob('*.ipynb')]
default_workflow_names = [
    file.replace('.ipynb', '')
    for file in available_workflows]

reports_to_exclude = ['COVID-Strain-Surveillance.ipynb']
reports_path = path_to_report_templates()
available_reports = [file.name for file in reports_path.glob('*.ipynb')
                     if not file.name in reports_to_exclude]
default_report_names = [
    file.replace('.ipynb', '')
    for file in available_reports]

diag_valid_params = [
    'metadata_path',
    'rt_partitions',
    'sampling_prop_partition_freq',
    'collection_date_field'
]

def _is_int(value):
    """Use casting to check if value can convert to an `int`."""
    try:
        int(value)
    except ValueError:
        return False
    else:
        return True


def _is_float(value):
    """Use casting to check if value can convert to a `float`."""
    try:
        float(value)
    except ValueError:
        return False
    else:
        return True

def _resolve_type(value):
    if value == "True":
        return True
    elif value == "False":
        return False
    elif value in ["None", "null"]:
        return None
    elif _is_int(value):
        return int(value)
    elif _is_float(value):
        return float(value)
    else:
        return value



@click.group(context_settings=dict(help_option_names=['-h', '--help']),
             epilog='See https://github.com/m-d-grunnill/BEAST_pype/wiki for further documentation.')
def beast_pype():
    pass


@beast_pype.command(context_settings=dict(help_option_names=['-h', '--help']))
@click.argument('workflow', required=1, type=click.Choice(default_workflow_names))
@click.argument('parameters', required=1, type=click.Path(exists=True, dir_okay=False, file_okay=True, readable=True))
@click.option('--kernel_name', '-k', default='beast_pype', type=str,
              help='Name of Jupyter python kernel_name to use when running workflow.\n' +
                   'This is also the name of the conda environment to use in phases 4 & ' +
                   'phase 2ii (as these Jupyter notebooks use the `bash` kernel_name).\n' +
                   'If not given "beast_pype" is used.'
              )
@click.option('--start_timeout', '-t', default=None, type=int,
              help='Time in seconds to wait for the kernel to start before raising an error when launching a jupyter notebook.\n' +
                   'If not given 60 seconds (1 minutes) is used.'
              )
def run_workflow(workflow,
                 parameters,
                 kernel_name,
                 start_timeout):
    """
    WORKFLOW: Workflow you wish to execute, or a path to jupyter notebook to be run  as a workflow.\n
    PARAMETERS: Path to YAML file containing parameters.
    """
    parameters = read_yaml_file(parameters)
    parameters['kernel_name'] = kernel_name
    workflow_save_name = f'{workflow}.ipynb'
    workflow = f'{workflows_path}/{workflow_save_name}'
    if 'specific_run_save_dir' not in parameters:
        parameters['specific_run_save_dir'] = datetime.now().strftime(
            "%Y-%m-%d_%H-%M-%S")
    os.makedirs(parameters['overall_save_dir'], exist_ok=True)
    save_dir = f"{parameters['overall_save_dir']}/{parameters['specific_run_save_dir']}"
    os.makedirs(save_dir)
    parameters['start_timeout'] = start_timeout
    execute_notebook(
        input_path=workflow,
        output_path=f"{save_dir}/{workflow_save_name}",
        parameters=parameters,
        kernel_name=kernel_name,
        progress_bar=True,
        nest_asyncio=True, start_timeout=start_timeout
    )

@beast_pype.command(context_settings=dict(help_option_names=['-h', '--help']))
@click.argument('report_template', required=1,
                type=click.Choice(default_report_names))
@click.argument('beast_outputs',
                 required=1,
                 type=click.Path(exists=True, dir_okay=True, file_okay=False, readable=True, writable=True))
@click.option('--parameters', '-p', nargs=2, multiple=True, help='Parameters to be passed to workflow.')
@click.option('--parameters_file', '-f', multiple=True, help='Path to YAML file containing parameters.')
@click.option('--kernel_name', '-k', default='beast_pype', type=str,
              help='Name of Jupyter python kernel_name to use when running diagnostic & report template notebooks.\n' +
                   'If not given "beast_pype" is used.'
              )
def diagnose_results(report_template,
                     beast_outputs,
                     parameters,
                     parameters_file,
                     kernel_name):
    """
    REPORT_TEMPLATE: Report template to use after diagnosing BEAST 2 outputs.\n
    BEAST_OUTPUTS: Path to directory containing BEAST 2 outputs to diagnose.
    """
    # Read in Parameters
    parameters_final = {}
    for name, value in parameters or []:
        parameters_final[name] = _resolve_type(value)
    for files in parameters_file or []:
        parameters_final.update(read_yaml_file(files) or {})

    if 'beast_xml_path' in parameters_final:
        beast_xml_path = parameters_final.pop('beast_xml_path')
    else:
        beast_xml_path = None
    for param, value in parameters_final.items():
        if param not in diag_valid_params:
            raise ValueError(f"Parameter {param} is not a valid parameter for use in the diagnostic workflow.")
    if report_template is None:
        raise ValueError('A value for --report_template is required for the diagnostic workflow.')
    if beast_outputs is None:
        raise ValueError('A value for --beast_outputs is required for the diagnostic workflow.')

    gen_beast_diagnostic_nb(
        beast_outputs=beast_outputs,
        report_template=report_template,
        beast_xml_path=beast_xml_path,
        kernel_name=kernel_name,
        **parameters_final)
